#!/usr/bin/bash
# Copyright 2020-2023 Tj <hacker@iam.tj>
# Licensed on the terms of the GNU Affero General Public Licence version 3.
# this script demonstrates fail-over and load-balancing of kernel managed
# ECMP + Anycast routing.
# The network is expected to be configure by the associated ecmp-anycast.sh and
# the file /tmp/ecmp.conf needs to exist detailing the address family and number
# of network namespaces being used.
# this script defaults to testing IPv4.
# give command-line parameter "-4" for IPv4 or "-6" for IPv6
#
## How to use
#  sudo ./ecmp-anycast.sh -6
#  sudo ./ecmp-anycast-test.sh
#
# the first command will report setting up 10 network namespaces
# it will write its config into /tmp/ecmp.conf which will be read by the -test script
# the test will run 10000 tests load-balanced across the 10 namespaces.
# At the end it will report how many requests each namespace received. The numbers should be
# almost equal if load-balancing worked.
# Sometimes there will be bind errors due to a (random) port already in use - this is to be
# expected in the test scenario on the same host.

set -e

echo > /tmp/sport.log

# associative array to track used ports
declare -A used
# pre-block the privileged ports
for n in $(seq 1 1023); do
  used[$n]=1
done

# IPv4 defaults
ipv4_ip=${ipv4_ip:-203.0.113.0}

# IPv6 defaults
ipv6_ip=${ipv6_ip:-fd00::1}

# session defaults
max=${max:-10}
port=${port:-1234}
message=${message:-Hello}
wt=${wt:-0}
protocol=${protocol:--u}
 
if [ -f /tmp/ecmp.conf ]; then
  read ip_family max < /tmp/ecmp.conf
fi

if [ -n "$1" ]; then
  ip_family="$1"
fi
ip_family=${ip_family:--4}

# operator inet family choice for session
case "$ip_family" in
  "-4")
    ip_anycast=$ipv4_ip
    ;;

  "-6")
    ip_anycast=$ipv6_ip
    ;;
esac
# send this many test messages
count=${count:-10000}

echo "Starting $count tests with $max namespaces using protocol $protocol address family $ip_family @ port $port" 

# create listener 'nc' processes in each namespace that also log requests
for nsnum in $(seq 1 $max); do
  ip netns exec ns$nsnum nc $ip_family -k $protocol -l $port  > /tmp/ns$nsnum.log &
done
jobs

while [ $count -gt 0 ]; do
  count=$((count -1))
  while /bin/true; do
		# send from a random port on the host; don't use a port already in use by this test
		# may still cause bind errors if a port used by some other process on the host is chosen
    sport=$(($RANDOM+$RANDOM))
    if [ -z ${used[$sport]} ]; then
      used[$sport]=1
      break
    fi
  done

	# log the port used
  echo $sport >> /tmp/sport.log
	# send a message to the anycast address
  echo "$message $count $sport " | nc $ip_family $protocol -w $wt -p $sport $ip_anycast $port || echo "error: $count $sport"
  if [ $((count %1000)) -eq 0 ]; then
    echo $count
  fi
done

# stop the 'nc' process in each namespace
for nsnum in $(seq 1 $max); do
  ip netns pids ns$nsnum | while read pid; do
    kill $pid
  done
done

# count and report how many requests each namespace received to demonstrate load-balancing
for f in /tmp/ns*.log; do 
  wc -l $f
  rm $f
done | sort -n
	
